unit untDOSMenu;

interface
uses
  Windows, SysUtils, ComServ, ComObj, ActiveX, ShlObj, ShellAPI, Dialogs, Registry, Graphics, Classes;

const
  GUID_TDOSMenuShellExt: TGUID = '{122498CD-2D6F-4356-BF8C-049ECAA61855}';

type
  TDOSMenuShellExt = class(TComObject, IShellExtInit, IContextMenu)
    protected
      function IShellExtInit.Initialize = SEInitialize;
      function SEInitialize(pidlFolder: PItemIDList; lpdobj: IDataObject; hKeyProgID: HKEY): HResult; stdcall;
      function QueryContextMenu(Menu: HMENU; indexMenu, idCmdFirst, idCmdLast, uflags: UINT): HResult; stdcall;
      function InvokeCommand(var lpici: TCMInvokeCommandInfo): HResult; stdcall;
      function GetCommandString(idCmd, uType: UINT; pwReserved: PUINT; pszName: LPSTR; cchMax: UINT): HResult; stdcall;
  end;

implementation

{ TDOSMenuShellExt }

type
  TDOSMenuShellExtFactory = class(TComObjectFactory)
    public
      procedure UpdateRegistry(Register: boolean); override;
  end;

var
  FFileName: array[0..MAX_PATH] of Char;
  hBmp1: TBitmap;
  hBmp2: TBitmap;

const
  CommandStrings: array[0..6] of string = ('/C Upx', '/K Dir', '/C Dir', '/K javac', '/C regsvr32', '/C regsvr32 /u', '/C start Notepad');

function GetConsoleOutput(const Command: String; var Output, Errors: TStringList): Boolean;
var
  StartupInfo: TStartupInfo;
  ProcessInfo: TProcessInformation;
  SecurityAttr: TSecurityAttributes;
  PipeOutputRead: THandle;
  PipeOutputWrite: THandle;
  PipeErrorsRead: THandle;
  PipeErrorsWrite: THandle;
  Succeed: Boolean;
  Buffer: array [0..255] of Char;
  NumberOfBytesRead: DWORD;
  Stream: TMemoryStream;
begin
  //Initialisierung ProcessInfo
  FillChar(ProcessInfo, SizeOf(TProcessInformation), 0);

  //Initialisierung SecurityAttr
  FillChar(SecurityAttr, SizeOf(TSecurityAttributes), 0);
  SecurityAttr.nLength := SizeOf(SecurityAttr);
  SecurityAttr.bInheritHandle := true;
  SecurityAttr.lpSecurityDescriptor := nil;

  //Pipes erzeugen
  CreatePipe(PipeOutputRead, PipeOutputWrite, @SecurityAttr, 0);
  CreatePipe(PipeErrorsRead, PipeErrorsWrite, @SecurityAttr, 0);

  //Initialisierung StartupInfo
  FillChar(StartupInfo, SizeOf(TStartupInfo), 0);
  StartupInfo.cb:=SizeOf(StartupInfo);
  StartupInfo.hStdInput := 0;
  StartupInfo.hStdOutput := PipeOutputWrite;
  StartupInfo.hStdError := PipeErrorsWrite;
  StartupInfo.wShowWindow := sw_Hide;
  StartupInfo.dwFlags := STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;

  if  CreateProcess(nil, PChar(command), nil, nil, true,
  CREATE_DEFAULT_ERROR_MODE or CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS, nil, nil,
  StartupInfo, ProcessInfo) then begin
    result:=true;
    //Write-Pipes schlieen
    CloseHandle(PipeOutputWrite);
    CloseHandle(PipeErrorsWrite);

    //Ausgabe Read-Pipe auslesen
    Stream := TMemoryStream.Create;
    try
      while true do begin
        succeed := ReadFile(PipeOutputRead, Buffer, 255, NumberOfBytesRead, nil);
        if not succeed then break;
        Stream.Write(Buffer, NumberOfBytesRead);
      end;
      Stream.Position := 0;
      Output.LoadFromStream(Stream);
    finally
      Stream.Free;
    end;
    CloseHandle(PipeOutputRead);

    //Fehler Read-Pipe auslesen
    Stream := TMemoryStream.Create;
    try
      while true do begin
        succeed := ReadFile(PipeErrorsRead, Buffer, 255, NumberOfBytesRead, nil);
        if not succeed then break;
        Stream.Write(Buffer, NumberOfBytesRead);
      end;
      Stream.Position := 0;
      Errors.LoadFromStream(Stream);
    finally
      Stream.Free;
    end;
    CloseHandle(PipeErrorsRead);

    WaitForSingleObject(ProcessInfo.hProcess, INFINITE);
    CloseHandle(ProcessInfo.hProcess);
  end
  else begin
    result:=false;
    CloseHandle(PipeOutputRead);
    CloseHandle(PipeOutputWrite);
    CloseHandle(PipeErrorsRead);
    CloseHandle(PipeErrorsWrite);
  end;
end;

procedure DoWork(index: Integer);
var
  tsi: TStartupInfo;
  tpi: TProcessInformation;
  cmdLine: array[0..512] of Char;
  command: string;
  ausgabe, fehler: TStringList;
begin
  FillChar(cmdLine, sizeof(cmdLine), 0);
  // Position des Befehlsinterpreter Command.com ermitteln
  GetEnvironmentVariable('ComSpec', cmdLine, sizeof(cmdLine) - 1);
  FillChar(tsi, sizeof(TStartupInfo), 0);
  tsi.cb := sizeof(TStartupInfo);
  tsi.wShowWindow := SW_SHOW;
  tsi.dwFlags     := STARTF_USESHOWWINDOW;

  if FileExists(FFileName) then
  begin
    command := StrPas(cmdLine) + ' ' + CommandStrings[index] + ' ' + ExtractFileName(FFileName);
    StrPCopy(cmdLine, command);
    CreateProcess(nil, cmdLine, nil, nil, True, 0, nil, PChar(ExtractFilePath(FFileName)), tsi, tpi);
  end
  else
  begin
    if index = 1 then
    begin
      command := StrPas(cmdLine) + ' ' + CommandStrings[index];
      StrPCopy(cmdLine, command);
      CreateProcess(nil, cmdLine, nil, nil, True, 0, nil, FFileName, tsi, tpi);
    end;
    if index = 2 then
    begin
      ausgabe := TStringList.Create;
      fehler  := TStringList.Create;
      try
        if GetConsoleOutput(StrPas(cmdLine) + ' /c dir ' + FFileName, ausgabe, fehler) then
          ShowMessage(ausgabe.Text);
      finally
        ausgabe.free;
        fehler.free;
      end;
    end;
  end;
end;

function TDOSMenuShellExt.GetCommandString(idCmd, uType: UINT;
  pwReserved: PUINT; pszName: LPSTR; cchMax: UINT): HResult;
begin
  if(idCmd = 0) then
  begin
    if(uType = GCS_HELPTEXT) then
      StrCopy(pszName, 'DOS-Konsole ffnen');
    Result := NOERROR;
  end
  else
    Result := E_INVALIDARG;
end;

function TDOSMenuShellExt.InvokeCommand(
  var lpici: TCMInvokeCommandInfo): HResult;
begin
  Result := E_FAIL;

  if (HiWord(Integer(lpici.lpVerb)) <> 0) then
    Exit;

  if LoWord(lpici.lpVerb) > Length(CommandStrings) then
    exit
  else
    DoWork(LoWord(lpici.lpVerb));

  Result := NOERROR;
end;

function TDOSMenuShellExt.QueryContextMenu(Menu: HMENU; indexMenu,
  idCmdFirst, idCmdLast, uflags: UINT): HResult;
var
  hMnu: HMENU;
  hMnu2: HMENU;
  vReg: TRegistry;
  Idx: Integer;
  mii: TMenuItemInfo;
  MenuCmdIdx: Integer;
  MenuIndex: Integer;
begin
  if ((uFlags and $0000000F) = CMF_NORMAL) or ((uFlags and CMF_EXPLORE) <> 0) or
     ((uFlags and CMF_VERBSONLY) <> 0) then
  begin
    hMnu := CreatePopupMenu();

    for MenuCmdIdx := 0 to Length(CommandStrings) - 1 do
    begin
      AppendMenu(hMnu, MF_STRING, idCmdFirst + MenuCmdIdx, PChar(CommandStrings[MenuCmdIdx]));
      if hBmp2.Handle <> 0 then
        SetMenuItemBitmaps(hMnu, MenuCmdIdx, MF_BYPOSITION, hBmp2.Handle, hBmp2.Handle);
    end;
    AppendMenu(hMnu, MF_SEPARATOR, 0, '' );

    mii.cbSize     := sizeof(TMenuItemInfo);
    mii.fMask      := MIIM_SUBMENU or MIIM_STRING or MIIM_ID;
    mii.wID        := idCmdFirst + Length(CommandStrings);
    mii.hSubMenu   := hMnu;
    mii.dwTypeData := PAnsiChar('DOSMen');
    InsertMenuItem(Menu, indexMenu, True, mii);
    if hBmp1.Handle <> 0 then
      SetMenuItemBitmaps(Menu, indexMenu, MF_BYPOSITION, hBmp1.Handle, hBmp1.Handle);

    mii.cbSize     := sizeof(TMenuItemInfo);
    mii.fMask      := MIIM_SUBMENU or MIIM_STRING or MIIM_ID;
    mii.wID        := idCmdFirst + Length(CommandStrings) + 1;
    mii.hSubMenu   := 0;
    mii.dwTypeData := PAnsiChar('DOSMen2');
    InsertMenuItem(Menu, indexMenu, True, mii);
    if hBmp1.Handle <> 0 then
      SetMenuItemBitmaps(Menu, indexMenu, MF_BYPOSITION, hBmp1.Handle, hBmp1.Handle);

    Result := MakeResult(SEVERITY_SUCCESS, FACILITY_NULL, Length(CommandStrings) + 1);
  end
  else
    Result := MakeResult(SEVERITY_SUCCESS, FACILITY_NULL, 0);
end;

function TDOSMenuShellExt.SEInitialize(pidlFolder: PItemIDList;
  lpdobj: IDataObject; hKeyProgID: HKEY): HResult;
var
  StgMedium: TStgMedium;
  FormatEtc: TFormatEtc;
begin
  if(lpdobj = nil) then
  begin
    Result := E_INVALIDARG;
    exit;
  end;

  with FormatEtc do
  begin
    cfFormat := CF_HDROP;
    ptd      := nil;
    dwAspect := DVASPECT_CONTENT;
    lindex   := -1;
    tymed    := TYMED_HGLOBAL;
  end;

  Result := lpdobj.GetData(FormatEtc, StgMedium);
  if Failed(Result) then
    exit;

  if DragQueryFile(StgMedium.hGlobal, $FFFFFFFF, nil, 0) = 1 then
    DragQueryFile(StgMedium.hGlobal, 0, FFileName, SizeOf(FFileName))
  else
    FFileName := '';

  ReleaseStgMedium(StgMedium);
  Result := NOERROR;
end;

{ TDOSMenuShellExtFactory }

procedure TDOSMenuShellExtFactory.UpdateRegistry(Register: boolean);
var
  ClassID: string;
begin
  if Register then
  begin
    inherited UpdateRegistry(Register);

    ClassID := GUIDToString(GUID_TDOSMenuShellExt);
    CreateRegKey('Folder\shellex', '', '');
    CreateRegKey('Folder\shellex\ContextMenuHandlers', '', '');
    CreateRegKey('Folder\shellex\ContextMenuHandlers\DOSMenu', '', ClassID);
    CreateRegKey('*\shellex', '', '');
    CreateRegKey('*\shellex\ContextMenuHandlers', '', '');
    CreateRegKey('*\shellex\ContextMenuHandlers\DOSMenu', '', ClassID);

    if (Win32Platform = VER_PLATFORM_WIN32_NT) then
      with TRegistry.Create do
        try
          RootKey := HKEY_LOCAL_MACHINE;
          OpenKey('SOFTWARE\Microsoft\Windows\CurrentVersion\Shell Extensions', True);
          OpenKey('Approved', True);
          WriteString(ClassID, 'DOSMenu');
        finally
          Free;
        end;
  end
  else
  begin
    DeleteRegKey('Folder\shellex\ContextMenuHandlers\DOSMenu');
    DeleteRegKey('*\shellex\ContextMenuHandlers\DOSMenu');

    inherited UpdateRegistry(Register);
  end;
end;

initialization
  TDOSMenuShellExtFactory.Create(ComServer, TDOSMenuShellExt,
                                 GUID_TDOSMenuShellExt, '',
                                 'DOSMenu', ciMultiInstance,
                                 tmApartment);
  hBmp1 := TBitmap.Create;
  hBmp1.LoadFromResourceName(hInstance, 'DOSMENU');
  hBmp2 := TBitmap.Create;
  hBmp2.LoadFromResourceName(hInstance, 'SUBMENU');
finalization
  hBmp1.Free;
  hBmp2.Free;
end.
